
\ Callist2.f 
\ Каллисто-2, черновик 7
\ Авторское право (c) Васильев Илья Владимирович, 18 ноября 2016, Москва
\ Перед вами язык программирования для "Электроники МК-161", развивающий идеи Каллисто Классик

\ Каллисто-2 распространяется под свободной лицензией GNU GPL версии 3 или более поздней.
\ Текст лицензии GPL на русском и английском языках располагается в архиве дистрибутива в каталоге gpl.
\ Каллисто-2, а также основаные на ней работы, запрещается распространять без исходного текста и текста лицензии.
\ Распространение приложений, написанных на Каллисто-2, и использование примеров кода на Каллисто-2,
\ размещённых в Руководстве, разрешается без ограничений.
\
\ Если вы несогласны с правами, которые предоставляет свободное программное обеспечение
\ и обязанностями, которые оно налагает, удалите Каллисто-2 со своих носителей.

\ Исходный текст предназначен для SP-Forth 4.20

\ == Регистры МК-161 ==
\ R2 RP        Указатель целочисленного стека возвратов, растёт вниз от 5090
\ R5           Временный регистр обработчика, может быть изменён NEXT, CALL, EXIT
\ R6 RI, IP    Указатель инструкций, указывает на токен в шитом коде
\ R7 W, WP     Указатель слов Форта, содержит CFA исполняемого слова
\              временный регистр обработчика
\ R8 JP        Указатель/регистр передачи управления NEXT,
\              временный регистр обработчика
\ R9 =NEXT     Адрес точки NEXT
\ RC TOS       Вершина стека данных
\
\ R9042 =9500  Индексный регистр памяти программ
\              Содержит адрес таблицы свёрток

: .s ( -- )  ." Stack: " DEPTH 0 DO I PICK . LOOP CR ;  \ Отладочная печать

\ -------------------------------------
\ Области памяти МК-161 и работа с ними
\ -------------------------------------

ONLY FORTH DEFINITIONS
CREATE prgm  10000 ALLOT         \ Память программ
prgm 10000 255 FILL              \ Заполняем FF

CREATE regb  8168 ALLOT          \ Регистровая память (байты)
regb 8168 ERASE                  \ Обнуляем

: save.mkp ( -- )                \ Сохранить память программ в файл Callist2.mkp
  HERE 00 C,                     \ Номер начальной страницы
  prgm 10000 00 DO DUP C@ C, 01 C, 1+ LOOP DROP
  S" Callist2.mkp" W/O ( BIN )
  CREATE-FILE ABORT" Can't create MKP"
  OVER 20001 02 PICK
  WRITE-FILE  ABORT" Can't write MKP"
  CLOSE-FILE  ABORT" Can't close MKP"
  -20001 ALLOT
;

: save.mkb ( -- )                \ Сохранить байтовые регистры в файл Callist2.mkb
  S" Callist2.mkb" W/O ( BIN )
  CREATE-FILE ABORT" Can't create MKB"
  1000 regb + 4096 02 PICK
  WRITE-FILE  ABORT" Can't write MKB"
  CLOSE-FILE  ABORT" Can't close MKB"
;

\ --------------------------------------------------------
\ X-компилятор исходного текста Каллисто-2 в память МК-161
\ --------------------------------------------------------

: 16* ( n1 -- n2 )  2* 2* 2* 2* ;                   \ Умножение на 16
: [+] + ;  : [-] - ;                                \ Сохраним операции с инструментальным стеком данных, как [+] и [-]
: getnumber ( "aaaa" -- )                           \ Считать следующее слово из входного потока, как число
    00 00 BL WORD COUNT >NUMBER 2DROP DROP ;

1000 VALUE bc                                       \ Номер текущего байтового регистра
: bhere ( -- a ) bc regb [+] ;                      \ Адрес текущего байтового регистра
: ballot ( n -- ) bc [+] TO bc ;                    \ Зарезервировать n байтовых регистров
: b, ( c -- ) bhere C! 01 ballot ;                  \ Скомпилировать байт c в байтовую память
: bxt, ( xt -- ) 100 /MOD b, b, ;                   \ Скомпилировать выполнимый токен xt в байтовую память
: .borg ( "aaaa" -- ) getnumber TO bc ;             \ Сделать байтовый регистр aaaa текущим
: b: ( "имя" -- ) CREATE bc , DOES> ( -- bc ) @ ;   \ Определить метку в байтовой памяти
: breg ( "имя" -- ) b: 01 ballot ;                  \ Дать имя очередному байтовому регистру

16 VALUE dc                                         \ Номер текущего десятичного регистра
: dallot ( n -- ) dc [+] TO dc ;                    \ Зарезервировать n десятичных регистров
: .dorg ( "aaaa" -- ) getnumber TO dc ;             \ Сделать десятичный регистр aaaa текущим
: d: ( "имя" -- ) CREATE dc , DOES> ( -- dc ) @ ;   \ Определить метку в десятичной памяти
: dreg ( "имя" -- ) d: 01 dallot ;                  \ Дать имя очередному десятичному регистру

00 VALUE pc                                         \ Счётчик команд
: phere ( -- a )  pc prgm [+] ;                     \ Адрес текущего места в памяти программ
: pallot ( n -- )  pc [+] TO pc ;                   \ Зарезервировать n шагов в памяти программ
: p, ( c -- )  phere C! 01 pallot ;                 \ Скомпилировать шаг c в память программ
: pbcd1, ( n -- ) 100 MOD 10 /MOD 16* [+] p, ;      \ Скомпилировать в шаг младшую половину n, как число bcd
: pbcd2, ( n -- ) 100 /MOD pbcd1, pbcd1, ;          \ Скомпилировать в два шага bcd-число n
: ptext" ( "text"" -- )                             \ Скомпилировать строку литер в память программ
    34 WORD COUNT SWAP phere 02 PICK MOVE pallot ;
: p: ( "имя" -- )  CREATE pc , DOES> ( -- pc ) @ ;  \ Определить метку в памяти программ
: .org ( "aaaa" -- )  getnumber TO pc ;             \ Сделать шаг по адресу aaaa текущим

00 VALUE sc                                         \ Номер свободной свёртки
: ps, ( pc -- ) 9500 prgm [+]                       \ Добавить адрес обработчика pc в таблицу свёрток
   sc 2* [+]  SWAP 256 /MOD
   02 PICK C!  SWAP 1+ C!  sc 1+ TO sc ;

\ ------------------------------------------------
\ X-компилятор входного языка "Электроники МК-161"
\ ------------------------------------------------

: 1mi ( cop -- )  CREATE C, DOES> ( -- ) C@ p, ;               \ Одношаговая команда
: 2mi ( cop -- )  CREATE C, DOES> ( "xx" -- ) C@ p,  pbcd1, ;  \ Двухшаговая команда
: 3mi ( cop -- )  CREATE C, DOES> ( "xx" -- ) C@ p,  pbcd2, ;  \ Трёхшаговая команда

VOCABULARY ЯМК

: code ( "имя" -- )                                 \ Начать x-определение примитива
    CREATE bc ,  sc b,  pc ps,  ALSO ЯМК  DOES> ( -- ) @ bxt, ;
: end-code ( -- ) ONLY FORTH ;                      \ Завершить x-определение примитива

: :: ( "имя" -- )                                   \ Начать x-определение высокого уровня
    CREATE bc ,  00 b,  DOES> ( -- ) @ bxt, ;       \ Сделать ;- для оптимизации хвоста
: ;; ( -- )  ;                                      \ Может компилировать exit, но пока пишем exit вручную

: ?pairs ( n1 n2 -- )                               \ Всё ли в порядке с операторными скобками?
    - ABORT" Conditionals aren't paired" ;
: ?far ( цель -- флаг )                             \ Проверить необходимость межстраничного перехода
    100 /  pc 1+ 100 /  = 0= ;

ONLY FORTH ALSO ЯМК DEFINITIONS
 07 CONSTANT x≠0                                    \ Коды условий для структур управления
 09 CONSTANT x≥0
0xC CONSTANT x<0
0xE CONSTANT x=0

\ Структуры управления
: begin ( -- цель 101 )  pc 101 ; IMMEDIATE         \ Отметить, куда будет переход назад, начало цикла
: until ( цель 101 условие -- ) SWAP 101 ?pairs     \ Условный переход назад, цикл "до"
    OVER ?far IF 0xF0 [+] p, pbcd2, ELSE 0x50 [+] p, pbcd1, THEN ; IMMEDIATE
: again ( цель 101 -- ) 101 ?pairs                  \ Безусловный переход назад, бесконечный цикл
    DUP ?far IF 0xF1 p, pbcd2, ELSE 0x51 p, pbcd1, THEN ; IMMEDIATE
: fl0 ( цель 101 -- ) 101 ?pairs                    \ Цикл по R0 назад, begin..fl0 = FL0/PFL0
    DUP ?far IF 0xFD p, pbcd2, ELSE 0x5D p, pbcd1, THEN ; IMMEDIATE
: fl1 ( цель 101 -- ) 101 ?pairs                    \ Цикл по R1 назад, begin..fl1 = FL1/PFL1
    DUP ?far IF 0xFB p, pbcd2, ELSE 0x5B p, pbcd1, THEN ; IMMEDIATE
: fl2 ( цель 101 -- ) 101 ?pairs                    \ Цикл по R2 назад, begin..fl2 = FL2/PFL2
    DUP ?far IF 0xF8 p, pbcd2, ELSE 0x58 p, pbcd1, THEN ; IMMEDIATE
: fl3 ( цель 101 -- ) 101 ?pairs                    \ Цикл по R3 назад, begin..fl3 = FL3/PFL3
    DUP ?far IF 0xFA p, pbcd2, ELSE 0x5A p, pbcd1, THEN ; IMMEDIATE

: pnum, ( число -- ) -1 SWAP                        \ Скомпилировать целое число в команды языка МК-161
    BEGIN 10 /MOD DUP 0= UNTIL DROP                 \ Сперва кладём все цифры в стек
    BEGIN p, DUP 0< UNTIL DROP                      \ Потом компилируем их в нужном порядке
;

   0 1mi 0        1 1mi 1          2 1mi 2          3 1mi 3         4 1mi 4        5 1mi 5        6 1mi 6        7 1mi 7
   8 1mi 8        9 1mi 9        0xA 1mi ,        0xB 1mi /-/     0xC 1mi ВП     0xD 1mi Cx     0xE 1mi В↑     0xF 1mi FВx
0x10 1mi +     0x11 1mi -       0x12 1mi ×       0x13 1mi ÷      0x14 1mi ↔     0x15 1mi F10ˣ  0x16 1mi Feˣ   0x17 1mi Flg
0x18 1mi Fln   0x19 1mi Farcsin 0x1A 1mi Farccos 0x1B 1mi Farctg 0x1C 1mi Fsin  0x1D 1mi Fcos  0x1E 1mi Ftg
0x20 1mi Fπ    0x21 1mi F√      0x22 1mi Fx²     0x23 1mi F1/x   0x24 1mi Fxʸ   0x25 1mi Fo    0x26 1mi КМ→Г  0x27 1mi K-
0x28 1mi КИПРГ 0x29 1mi К÷      0x2A 1mi КМС→Г
0x30 1mi КГ→МС 0x31 1mi К|x|    0x32 1mi КЗН     0x33 1mi КГ→М   0x34 1mi К[x]  0x35 1mi К{x}  0x36 1mi Кmax  0x37 1mi КAND
0x38 1mi КOR   0x39 1mi КXOR    0x3A 1mi КNOT    0x3B 1mi КСЧ
0x40 1mi П0    0x41 1mi П1      0x42 1mi П2      0x43 1mi П3     0x44 1mi П4    0x45 1mi П5    0x46 1mi П6    0x47 1mi П7
0x48 1mi П8    0x49 1mi П9      0x4A 1mi ПА      0x4B 1mi ПВ     0x4C 1mi ПС    0x4D 1mi ПД    0x4E 1mi ПЕ    0x4F 2mi РП
0x50 1mi С/П   0x51 2mi БП      0x52 1mi В/О     0x53 2mi ПП     0x54 1mi КНОП  0x55 1mi КЭКР  0x56 1mi КГРФ \ 0x57 4mi Fx≠0

0x60 1mi ИП0   0x61 1mi ИП1     0x62 1mi ИП2     0x63 1mi ИП3    0x64 1mi ИП4   0x65 1mi ИП5   0x66 1mi ИП6   0x67 1mi ИП7
0x68 1mi ИП8   0x69 1mi ИП9     0x6A 1mi ИПА     0x6B 1mi ИПВ    0x6C 1mi ИПС   0x6D 1mi ИПД   0x6E 1mi ИПЕ   0x6F 2mi РИП
0x70 1mi Кx≠00 0x71 1mi Кx≠01   0x72 1mi Кx≠02   0x73 1mi Кx≠03  0x74 1mi Кx≠04 0x75 1mi Кx≠05 0x76 1mi Кx≠06 0x77 1mi Кx≠07
0x78 1mi Кx≠08 0x79 1mi Кx≠09   0x7A 1mi Кx≠0А   0x7B 1mi Кx≠0В  0x7C 1mi Кx≠0С 0x7D 1mi Кx≠0Д 0x7E 1mi Кx≠0Е 0x7F 2mi РКx≠0
0x80 1mi КБП0  0x81 1mi КБП1    0x82 1mi КБП2    0x83 1mi КБП3   0x84 1mi КБП4  0x85 1mi КБП5  0x86 1mi КБП6  0x87 1mi КБП7
0x88 1mi КБП8  0x89 1mi КБП9    0x8A 1mi КБПА    0x8B 1mi КБПВ   0x8C 1mi КБПС  0x8D 1mi КБПД  0x8E 1mi КБПЕ  0x8F 2mi РКБП
0x90 1mi Кx≥00 0x91 1mi Кx≥01   0x92 1mi Кx≥02   0x93 1mi Кx≥03  0x94 1mi Кx≥04 0x95 1mi Кx≥05 0x96 1mi Кx≥06 0x97 1mi Кx≥07
0x98 1mi Кx≥08 0x99 1mi Кx≥09   0x9A 1mi Кx≥0А   0x9B 1mi Кx≥0В  0x9C 1mi Кx≥0С 0x9D 1mi Кx≥0Д 0x9E 1mi Кx≥0Е 0x9F 2mi РКx≥0
0xA0 1mi КПП0  0xA1 1mi КПП1    0xA2 1mi КПП2    0xA3 1mi КПП3   0xA4 1mi КПП4  0xA5 1mi КПП5  0xA6 1mi КПП6  0xA7 1mi КПП7
0xA8 1mi КПП8  0xA9 1mi КПП9    0xAA 1mi КППА    0xAB 1mi КППВ   0xAC 1mi КППС  0xAD 1mi КППД  0xAE 1mi КППЕ  0xAF 2mi РКПП
0xB0 1mi КП0   0xB1 1mi КП1     0xB2 1mi КП2     0xB3 1mi КП3    0xB4 1mi КП4   0xB5 1mi КП5   0xB6 1mi КП6   0xB7 1mi КП7
0xB8 1mi КП8   0xB9 1mi КП9     0xBA 1mi КПА     0xBB 1mi КПВ    0xBC 1mi КПС   0xBD 1mi КПД   0xBE 1mi КПЕ   0xBF 2mi РКП
0xC0 1mi Кx<00 0xC1 1mi Кx<01   0xC2 1mi Кx<02   0xC3 1mi Кx<03  0xC4 1mi Кx<04 0xC5 1mi Кx<05 0xC6 1mi Кx<06 0xC7 1mi Кx<07
0xC8 1mi Кx<08 0xC9 1mi Кx<09   0xCA 1mi Кx<0А   0xCB 1mi Кx<0В  0xCC 1mi Кx<0С 0xCD 1mi Кx<0Д 0xCE 1mi Кx<0Е 0xCF 2mi РКx<0
0xD0 1mi КИП0  0xD1 1mi КИП1    0xD2 1mi КИП2    0xD3 1mi КИП3   0xD4 1mi КИП4  0xD5 1mi КИП5  0xD6 1mi КИП6  0xD7 1mi КИП7
0xD8 1mi КИП8  0xD9 1mi КИП9    0xDA 1mi КИПА    0xDB 1mi КИПВ   0xDC 1mi КИПС  0xDD 1mi КИПД  0xDE 1mi КИПЕ  0xDF 2mi РКИП
0xE0 1mi Кx=00 0xE1 1mi Кx=01   0xE2 1mi Кx=02   0xE3 1mi Кx=03  0xE4 1mi Кx=04 0xE5 1mi Кx=05 0xE6 1mi Кx=06 0xE7 1mi Кx=07
0xE8 1mi Кx=08 0xE9 1mi Кx=09   0xEA 1mi Кx=0А   0xEB 1mi Кx=0В  0xEC 1mi Кx=0С 0xED 1mi Кx=0Д 0xEE 1mi Кx=0Е 0xEF 2mi РКx=0
               0xF1 3mi РБП     0xF2 1mi РРВ/О   0xF3 3mi РПП    0xF4 3mi РРП                  0xF6 3mi РРИП  0xF7 3mi Рx≠0
0xF8 3mi РFL2  0xF9 3mi Рx≥0    0xFA 3mi РFL3    0xFB 3mi РFL1   0xFC 3mi Рx<0  0xFD 3mi РFL0  0xFE 3mi Рx=0

\ --------------------------------------
\ Каллисто-2: Десятичные регистры МК-161
\ --------------------------------------

\ ------------------------------------
\ Каллисто-2: Байтовые регистры МК-161
\ ------------------------------------
b: bufKbd  08 ballot  \ Буфер клавиатуры размером в 8 байт
breg kbdHead          \ Голова буфера клавиатуры: 0..7
breg kbdTail          \ Хвост буфера клавиатуры: 0..7
breg bKey             \ Код нажатой клавиши (или 0)

\ ----------------------------------
\ Каллисто-2: Память программ МК-161
\ ----------------------------------

Cx                           \ === Запускаем шитый код ===
9 5 0 0                      \ Адрес таблицы свёрток
9042 РРП                     \ Индексный регистр памяти программ
5 0 9 0 П2                   \ Инициализация стека возвратов
1 0 9 9 П6                   \ RI := Начало шитого кода - 1
pc 03 [+] pnum,              \ Здесь мы подразумевам, что адрес NEXT меньше 100
П9                           \ Занесём в R9 адрес NEXT


p: NEXT                      \ Адресный интерпретатор, выполнить следующий токен в шитом коде
КИП6 ВП 2 КИП6 + П7          \ W := REGW[RI++], считываем CFA в R7
КИП7 9210 РРП                \ Находим по CFA адрес обработчика
П8 КБП8                      \ Передаём на него управление

p: CALL                      \ Обработчик слов высокого уровня
ИП6 В↑ ВП 2 /-/ К[x] КП2     \ Сохраняем старшую половину RI
ВП 2 - КП2                   \ Сохраняем младшую половину RI
ИП7 П6                       \ RI := PFA - 1 = CFA
КБП9                         \ Для быстродействия CALL или EXIT надо склеить с NEXT

CALL ps,                     \ Обработчик 000: CALL

begin
p: ChPutBell                 \ Издать короткий звуковой сигнал
  4 4 0 В↑ 1 0
  9052 РРП
  РРВ/О                      \ Может завершать прерывание клавиатуры или таймера

p: Timer1Int                 \ Прерывание: таймер 1
  2 0 9051 РРП               \ 200 мс, 5 раз/секунду

p: KeyAdd
  kbdHead РРИП               \ Голова очереди клавиатуры
  kbdTail РРИП ПА 1 + 7 КAND \ Сдвинуть хвост на 1 байт
  - x≠0                      \ Если совпадёт с головой, недовольно пискнуть
until

FВx kbdTail РРП              \ Сдвинуть хвост
bufKbd pnum, ИПА + ПА        \ RA: куда записать нажатую клавишу
bKey РРИП КПА                \ Записать в очередь клавиатуры код нажатой клавиши
РРВ/О                        \ Возврат из прерывания

p: tblKeys                   \ Таблица кодов клавиш
ptext" 0123456789,. "
08 p, 24 p,
ptext" +-"
179 p, 176 p, 183 p,         \ × ÷ ↔
ptext" F?P():;@![]P"
27 p, 13 p,                  \ ВЫХОД ВВОД
26 p, 27 p, 24 p, 25 p,

p: KeyPress                  \ Прерывание: нажатие на кнопку клавиатуры
9029 РРИП                    \ Чтение кода нажатой кнопки
tblKeys pnum, + КИПРГ        \ Перевести в код литеры
bKey РРП                     \ Записать в bKey
5 0 9051 РРП                 \ Задержка 500 мс (1/2 с) перед повтором
3 9151 РРП                   \ Регистр разрешения аппаратных прерываний: окончание счёта таймера 1: прерывание 3
KeyAdd РБП                   \ Добавить клавишу в очередь

p: KeyRelease                \ Прерывание: отпускание кнопки клавиатуры
Cx bKey РРП                  \ Сбрасываем код нажатой клавиши
9151 РРП                     \ Запретить аппаратное прерывание от таймера 1
9051 РРП                     \ Обнуляем таймер 1
РРВ/О                        \ Возврат из прерывания

p: CsrOn
p: CsrOff
9001 РРИП +
3 9001 РРП
8 В↑ 9013 РРП
КГРФ
FВx 9001 РРП В/О

p: ChGet                     \ Ждём нажатия клавиши и берём её код из очереди клавиатуры
CsrOn РПП
begin
  kbdTail РРИП kbdHead РРИП - x≠0
until
bufKbd pnum, FВx + ПА КИПА ПА
FВx 1 + 7 КAND kbdHead РРП
CsrOff РПП
ИПА В/О

p: ChPutLF                   \ Перевод строки, пока без прокрутки
9000 РРИП ↔ 8 + ↔ 9000 РРП В/О

p: ChPut2                    \ Вывод литеры из RA на графический экран
ИПА 9020 РРП В/О

p: ChPut                     \ Терминальный ввод/вывод
ПА
9000 РРИП
1 2 1 - ChPut2 Рx≥0
Cx 9000 РРП
ChPutLF РПП
ChPut2 РБП


ONLY FORTH DEFINITIONS

code exit                    \ === Примитив exit ===
  ИП2 П8 П5  2 + П2          \ RP += 2
  КИП5 ВП 2 КИП8 + П6        \ RPOP RI
  КБП9                       \ NEXT
end-code

\ Основной код инициализации Каллисто-2
code init                    \ === Примитив init ===
  1 9148 РРП                 \ Сброс регистров прерываний
  KeyPress pnum, В↑ 1 2 7    \ Вектор KeyPress, приоритет 127
  9131 РРП                   \  → регистр настройки прерывания 1
  1 9153 РРП                 \ Регистр разрешения аппаратных прерываний: нажатие на кнопку клавиатуры: прерывание 1
  KeyRelease pnum, В↑ 1 2 8  \ Вектор KeyRelease, приоритет 128
  9132 РРП                   \  → регистр настройки прерывания 2
  2 9157 РРП                 \ Регистр разрешения аппаратных прерываний: отпускание кнопки клавиатуры: прерывание 2
  Timer1Int pnum, В↑ 1 2 6   \ Вектор Timer1Int, приоритет 126
  9133 РРП                   \  → регистр настройки прерывания 3
  1 9130 РРП                 \ Общее разрешение прерываний: прерывания разрешены
  2 9010 РРП                 \ Инициализация графического экрана
  КБП9                       \ NEXT
end-code

code key  ChGet РПП ПС КБП9  end-code
code emit  ИПС ChPut РПП КБП9  end-code

code test                    \ === Примитив test ===
  begin
    ChGet РПП ПС
    ChPut РПП КГРФ
    ИПС 2 7 - x=0
  until
  КБП9
end-code

code bye                     \ === Примитив bye ===
  С/П                        \ Выход в Калькулятор
  КБП9
end-code

:: testAAA                   \ Определения высокого уровня, пока для проверки
     test bye
     exit
;;

\ ------------------------------------
\ Каллисто-2: Байтовые регистры МК-161
\ ------------------------------------

.borg 1100                   \ === Автозапускающийся шитый код ===
init
testAAA
bye

.( Compiled! )

\ Запишем Каллисто-2 на диск
save.mkp save.mkb

.( Saved!) CR

\EOF
